#include <xen/config.h>
#include <xen/init.h>
#include <xen/errno.h>
#include <xen/event.h>
#include <xen/irq.h>
#include <xen/smp.h>
#include <xen/list.h>
#include <xen/softirq.h>
#include <xen/perfc.h>
#include <xen/sched.h>
#include <xen/keyhandler.h>
#include <xen/compat.h>
#include <xen/iocap.h>
#include <xen/iommu.h>
#include <xen/bitops.h>

#include <asm/current.h>
#include <asm/irq.h>
#include <asm/flushtlb.h>

#include <public/physdev.h>

static spinlock_t irq_lock = SPIN_LOCK_UNLOCKED;

/*
 * Generic no controller code
 */

static void enable_none(unsigned int irq) { }
static unsigned int startup_none(unsigned int irq) { return 0; }
static void disable_none(unsigned int irq) { }
static void ack_none(unsigned int irq)
{
/*
 * 'what should we do if we get a hw irq event on an illegal vector'.
 * each architecture has to answer this themselves, it doesn't deserve
 * a generic callback i think.
 */
	// printk(KERN_ERR "Unexpected irq vector 0x%x on CPU %u!\n", irq, smp_processor_id());
}

/* startup is the same as "enable", shutdown is same as "disable" */
#define shutdown_none   disable_none
#define end_none        enable_none

struct hw_interrupt_type no_irq_type = {
	"none",
	startup_none,
	shutdown_none,
	enable_none,
	disable_none,
	ack_none,
	end_none
};

irq_desc_t irq_desc[NR_VECTORS] = {
	[0 ... NR_VECTORS - 1] = {
		.status = IRQ_DISABLED,
		.handler = &no_irq_type,
		.lock = SPIN_LOCK_UNLOCKED
	}
};

/*
 * Public API
 */

int init_IRQ(void)
{
	/* Initialize the irq descriptors */
	// write_64bit_cp0_eimr(phnx_irq_mask);

	init_ipi();

	return 0;
}

int pirq_guest_bind(struct vcpu *v, int irq, int will_share)
{
	return 0;
}

void pirq_guest_unbind(struct domain *d, int irq)
{

}

int setup_irq_vector(unsigned int vector, struct irqaction *new)
{
    irq_desc_t *desc = &irq_desc[vector];
    unsigned long flags;

    spin_lock_irqsave(&irq_lock, flags);
    write_64bit_cp0_eimr(read_64bit_cp0_eimr() | (1ULL << vector));
    spin_unlock_irqrestore(&irq_lock, flags);

    spin_lock_irqsave(&desc->lock,flags);

    if (desc->action != NULL) {
        spin_unlock_irqrestore(&desc->lock,flags);
        return -EBUSY;
    }

    desc->action  = new;
    desc->depth   = 0;
    desc->status &= ~IRQ_DISABLED;
    desc->handler->startup(vector);

    spin_unlock_irqrestore(&desc->lock,flags);

    return 0;
}

/*
 * The interface defined by Xen's setup_irq_vector() is 
 * insufficent. For example, it is not possible to
 * specify that an IRQ/interrupt be treated as a per-cpu
 * IRQ/interrupt, that should not be handled by cpus to
 * which it is not destined. Although the same could be
 * specified thorugh "irqflags" argument of request_irq_vector(),
 * irqflags do not trickle down to setup_irq_vector(),
 * since there is no provision for passing flags.
 * 
 * We are duplicating setup_irq_vector() code as reusing
 * the code by making a call to setup_irq_vector() and then
 * use irqflags separately is prone to race conditions
 */
static int __setup_irq_vector(unsigned int vector, struct irqaction *new, unsigned int irqflags)
{
    irq_desc_t *desc = &irq_desc[vector];
    unsigned long flags;

    spin_lock_irqsave(&irq_lock, flags);
    write_64bit_cp0_eimr(read_64bit_cp0_eimr() | (1ULL << vector));
    spin_unlock_irqrestore(&irq_lock, flags);

    spin_lock_irqsave(&desc->lock,flags);

    if (desc->action != NULL) {
        spin_unlock_irqrestore(&desc->lock,flags);
        return -EBUSY;
    }

    desc->action  = new;
    desc->depth   = 0;
    desc->status &= ~IRQ_DISABLED;
    desc->status |= irqflags;
    desc->handler->startup(vector);

    spin_unlock_irqrestore(&desc->lock,flags);

    return 0;
}

int request_irq_vector(unsigned int vector, 
        void (*handler)(int, void *, struct cpu_user_regs *),
        unsigned long irqflags, const char *devname, void *dev_id)
{
    struct irqaction * action; 
    int retval; 

    /*
     * Sanity-check: shared interrupts must pass in a real dev-ID, 
     * otherwise we'll have trouble later trying to figure out
     * which interrupt is which (messes up the interrupt freeing logic etc).
     *                          
	 */
    if (vector >= NR_VECTORS)
        return -EINVAL;
    if (!handler)
        return -EINVAL;

    action = xmalloc(struct irqaction);
    if (!action)
        return -ENOMEM;

    action->handler = handler;
    action->name = devname;
    action->dev_id = dev_id; 

    retval = __setup_irq_vector(vector, action, (unsigned int)irqflags);
    if (retval)
        xfree(action);

    return retval; 
}

void release_irq_vector(unsigned int vec)
{
    unsigned long flags;
    irq_desc_t *desc;

    if (vec >= NR_VECTORS)
        return;

    desc = &irq_desc[vec];

    spin_lock_irqsave(&desc->lock, flags); 

    spin_lock_irqsave(&irq_lock, flags);
    write_64bit_cp0_eimr(read_64bit_cp0_eimr() & ~(1ULL << vec));
    spin_unlock_irqrestore(&irq_lock, flags);

    desc->action = NULL; 
    desc->depth = 1;
    desc->status |= IRQ_DISABLED;
    desc->handler->shutdown(vec);
    spin_unlock_irqrestore(&desc->lock, flags); 

    while (desc->status & IRQ_INPROGRESS)
        cpu_relax();
}

fastcall unsigned int __do_IRQ(unsigned int irq, struct cpu_user_regs *regs)
{
	irq_desc_t *desc = irq_desc + irq;
	struct irqaction * action;
	unsigned int status;

	/* Debug hooks to catch unhandled interrupts from Guest OS! */
	if (irq > 63) {
		printk("[%s]: cpu-%d bad irq=%d\n", __func__, processor_id(), irq);
		cpu_halt();
	}
	else {
		if (irq != 4 && irq != 7 && irq != 63) {
			printk("[%s]: cpu-%d bad irq=%d\n", __func__, processor_id(), irq);
			cpu_halt();
		}
	}

	if (likely(desc->status & IRQ_PER_CPU)) {
		/*
		 * No locking required for CPU-local interrupts:
		 */
		desc->handler->ack(irq);
		local_irq_enable();
		desc->action->handler(irq, desc->action->dev_id, regs);
		local_irq_disable();
		desc->handler->end(irq);
		return 1;
	}

	spin_lock(&desc->lock);

#if 0
	if (desc->status & IRQ_GUEST) {
		/*
		 * are interrupts to guest domains bounced here ?
		 */
		__do_IRQ_guest(irq);
		spin_unlock(&desc->lock);
		return 1;
	}
#endif

	desc->handler->ack(irq);
	status = desc->status & ~IRQ_REPLAY;
	status |= IRQ_PENDING; /* we _want_ to handle it */

	/*
	 * If the IRQ is disabled for whatever reason, we cannot
	 * use the action we have.
	 */
	action = NULL;
	if (likely(!(status & (IRQ_DISABLED | IRQ_INPROGRESS)))) {
		action = desc->action;
		status &= ~IRQ_PENDING; /* we commit to handling */
		status |= IRQ_INPROGRESS; /* we are handling it */
	}
	desc->status = status;

	/*
	 * If there is no IRQ handler or it was disabled, exit early.
	 * Since we set PENDING, if another processor is handling
	 * a different instance of this same irq, the other processor
	 * will take care of it.
	 */
	if (unlikely(!action))
		goto out;

	/*
	 * Edge triggered interrupts need to remember
	 * pending events.
	 * This applies to any hw interrupts that allow a second
	 * instance of the same irq to arrive while we are in do_IRQ
	 * or in the handler. But the code here only handles the _second_
	 * instance of the irq, not the third or fourth. So it is mostly
	 * useful for irq hardware that does not mask cleanly in an
	 * SMP environment.
	 */
	for (;;) {
		spin_unlock_irq(&desc->lock);
		action->handler(irq, action->dev_id, regs);
		spin_lock_irq(&desc->lock);

		if (likely(!(desc->status & IRQ_PENDING)))
			break;

		desc->status &= ~IRQ_PENDING;
	}
	desc->status &= ~IRQ_INPROGRESS;

out:
	/*
	 * The ->end() handler has to deal with interrupts which got
	 * disabled while the handler was running.
	 */
	desc->handler->end(irq);
	spin_unlock(&desc->lock);

	return 1;
}

asmlinkage void plat_irq_dispatch(void)
{
	uint64_t eirr;
	int i = 0;
	struct pt_regs *pt_regs = current_thread_info()->regs;
	eirr = read_64bit_cp0_eirr() & read_64bit_cp0_eimr();

	if (!eirr)
		return;

	i = __ilog2(eirr);
	if (i == -1) {
		printk("no interrupt !!\n");
		return;
	}

	/* ack eirr */
	write_64bit_cp0_eirr(1ULL << i);
	irq_enter();
	__do_IRQ(i, pt_regs);
	irq_exit();
	return;
}

void irq_exit(void)
{
	local_irq_count(smp_processor_id())--;
	
	if (!in_irq()) {
		local_irq_enable();
		do_softirq();
		local_irq_disable();
	}
}
