#include <xen/config.h>
#include <xen/types.h>
#include <xen/kernel.h>
#include <xen/string.h>
#include <xen/console.h>

#include <asm/bootinfo.h>
#include <asm/sgidefs.h>
#include <asm/mapops.h>

/*
 *      _____________        ____              _______
 *     |             |      |    |            |       |
 * ____|             |______|    |____________|       |_______
 *        ______________                __________________
 *       |              |              |                  |
 * ______|              |______________|                  |___
 *
 *
 * Caution: The following functions are meant to operate on
 *          maps of the same type
 */

/*
 * FIXME: type field needs to be populated during merge
 */
struct addrmap *merge_maps(struct addrmap *map1, struct addrmap *map2, struct addrmap *scratch)
{
	int i, j, k;
	uint64_t prev_end = 0, next_end, next_start;
	struct addrmap_entry *list_one, *list_two, *merged_list;
   
	list_one = &map1->map[0]; list_two = &map2->map[0];
	merged_list = &scratch->map[0];
	list_one[map1->nr_map].addr = list_two[map2->nr_map].addr = -1;

	for (i = 0, j = 0, k = 0; (i < map1->nr_map) || (j < map2->nr_map);) {
		if (list_one[i].addr < list_two[j].addr) {
			next_start = list_one[i].addr;
			next_end = list_one[i].addr + list_one[i].size;
			++i;
		}
		else if (list_one[i].addr > list_two[j].addr) {
			next_start = list_two[j].addr;
			next_end = list_two[j].addr + list_two[j].size;
			++j;
		}
		else {
			next_start = list_one[i].addr;
			next_end = list_one[i].addr + max(list_one[i].size, list_two[j].size);
			++i; ++j;
		}
		if (next_end <= prev_end)
			continue;
		if (next_start < prev_end)
			next_start = prev_end;
		merged_list[k].addr = next_start;
		merged_list[k].size = next_end - next_start;
		prev_end = next_end;
		++k;
	}
	scratch->nr_map = k;

	return scratch;
}

struct addrmap *subtract_map(struct addrmap *from, struct addrmap *delete, struct addrmap *scratch)
{
	int i, j, k;
	uint64_t next_size;
	struct addrmap_entry *from_list, *delete_list, *pruned_list;

	from_list = &from->map[0]; delete_list = &delete->map[0];
	pruned_list = &scratch->map[0];

	from_list[from->nr_map].addr = delete_list[delete->nr_map].addr = -1;
	from_list[from->nr_map].size = delete_list[delete->nr_map].size = 0;
	from_list[from->nr_map].type = from_list[0].type;

	for (i = 0, j = 0, k = 0; (i < from->nr_map) || (j < delete->nr_map);) {
		if (from_list[i].addr < delete_list[j].addr) {
			next_size = min(from_list[i].addr + from_list[i].size, delete_list[j].addr) -
				from_list[i].addr;
			pruned_list[k].addr = from_list[i].addr;
			if (from_list[i].size > next_size) {
				from_list[i].addr += next_size;
				from_list[i].size -= next_size;
			}
			else {
				++i;
			}
			pruned_list[k].size = next_size;
			pruned_list[k].type = from_list[i].type;
			++k;
		}
		else if (from_list[i].addr > delete_list[j].addr) {
			next_size = min(delete_list[j].addr + delete_list[j].size, from_list[i].addr) -
				delete_list[j].addr;
			if (delete_list[j].size > next_size) {
				delete_list[j].addr += next_size;
				delete_list[j].size -= next_size;
			}
			else
				++j;
		}
		else {
			next_size = min(from_list[i].size, delete_list[j].size);
			if (from_list[i].size > next_size) {
				from_list[i].addr += next_size;
				from_list[i].size -= next_size;
				++j;
			}
			else if (delete_list[j].size > next_size) {
				delete_list[j].addr += next_size;
				delete_list[j].size -= next_size;
				++i;
			}
			else {
				 ++i; ++j;
			}
		}
	}

	scratch->nr_map = k;

	return scratch;
}

void coalesce_map(struct addrmap *map)
{
	int i, k;
	struct addrmap_entry *list = &(map->map[0]);

	for (i = 0, k = -1; (i < map->nr_map) && (list[i].size == 0); ++i)
		;
	if (i < map->nr_map)
		list[++k] = list[i];

	for (++i; i < map->nr_map; ++i) {
		if (list[i].size == 0)
			continue;
		if (list[k].addr + list[k].size >= list[i].addr)
			list[k].size = list[i].addr + list[i].size - list[k].addr;			
		else
			list[++k] = list[i];
	}
	map->nr_map = k + 1;
}

void sort_map(struct addrmap *map)
{
	int i,j;
	struct addrmap_entry *list, tmp;
	
	list = &(map->map[0]);
	for (i = 0; i < map->nr_map; i++) {
		for (j = i; j < map->nr_map; j++) {
			if (list[i].addr > list[j].addr) {
				tmp = list[i];
				list[i] = list[j];
				list[j] = tmp;
			}
		}
	}
}

int serialize_map(char *dst, struct addrmap *addrmap)
{
	memcpy(dst, addrmap, sizeof(struct addrmap));

	return sizeof(struct addrmap);
}

int deserialize_map(char *src, struct addrmap *addrmap)
{
	memcpy(src, addrmap, sizeof(struct addrmap));

	return (sizeof(struct addrmap));
}

void print_map(struct addrmap *addrmap)
{
	int i;

	printk("(%s): nr_map = %d\n", __FUNCTION__, addrmap->nr_map);
	for (i = 0; i < addrmap->nr_map; i++) {
		printk("\t%010llx --> %010llx \n",
			   (unsigned long long)addrmap->map[i].addr,
			   (unsigned long long)(addrmap->map[i].addr + addrmap->map[i].size - 1));
	}
}

void delete_map(struct addrmap *map)
{
	map->nr_map = 0;
}

int serialize_strings(char *dst, char **s)
{
	int narg;
	char *c, *_dst = dst;
	
	for (narg = 0; *(s + narg); ++narg)
		;
	memcpy(dst, &narg, sizeof(int));
	dst += sizeof(int);
	for (; *s; ++s) {
		for (c = *s; *c; ++c, ++dst)
			*dst = *c;
		*dst++ = '\0';
	}

	return (dst - _dst);
}


int deserialize_strings(char **dst, char *src)
{
	int narg, i;
	char *_src = src;
	
	memcpy(&narg, src, sizeof(int)); 
	src += sizeof(int);
	for (i = 0; i < narg; ++i) {
		dst[i] = src;
		src += strlen(src) + 1;
	}
	dst[i] = (void *)0;

	return (src - _src);
}

int add_addrmap_entry(struct addrmap *memmap, uint64_t addr, uint64_t size, uint32_t type)
{
	if (memmap->nr_map == MEMMAP_MAX_ENTRIES)
		return -ENOMEM;

	memmap->map[memmap->nr_map].addr = addr;
	memmap->map[memmap->nr_map].size = size;
	memmap->map[memmap->nr_map].type = type;

	memmap->nr_map++;

	return 0;
}
