#include <xen/config.h>
#include <xen/threads.h>
#include <xen/init.h>
#include <xen/linkage.h>
#include <xen/types.h>
#include <xen/sched.h>
#include <xen/smp.h>
#include <xen/spinlock.h>
#include <xen/serial.h>
#include <xen/xmalloc.h>
#include <xen/string.h>
#include <xen/serial.h>
#include <xen/keyhandler.h>
#include <xen/vga.h>
#include <xen/console.h>
#include <xen/lib.h>
#include <xen/list.h>
#include <xen/softirq.h>
#include <xen/mm.h>

#include <xen/domain.h>
#include <xen/debug.h>
#include <xsm/xsm.h>
#include <xsm/acm/acm_hooks.h>

#include <public/version.h>

#include <asm/cpu.h>
#include <asm/cpu-info.h>
#include <asm/netl/iomap.h>
#include <asm/netl/pic.h>
#include <asm/netl/uart.h>
#include <asm/netl/asmmacro.h>
#include <asm/bootinfo.h>
#include <asm/pfn.h>
#include <asm/netl/cache.h>
#include <asm/traps.h>
#include <asm/thread_info.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/console.h>
#include <asm/mapops.h>

#define swap(x, y) \
({ \
__typeof__(x) z; \
z = (x); \
(x) = (y); \
(y) = (z); \
})

/*
 * A few of the key data structures for Xen initialization
 */
struct cpuinfo_mips cpu_data[NR_CPUS] __read_mostly;
unsigned long xensp[NR_CPUS];
struct task_struct init_task;

static struct addrmap mapone, maptwo, mapthree;
struct addrmap *drammap_p = &mapone;
struct addrmap *paddrmap_p = &maptwo;
struct addrmap *scratchmap_p = &mapthree;

#define CHUNK1_START    0x400000
#define CHUNK1_SIZE    (0x1000000 - 0x400000)
#define CHUNK2_START   0x50000000
#define CHUNK2_SIZE    0x20000000

static struct addrmap map4 = {2,
			      {
				      {CHUNK1_START, CHUNK1_SIZE, RAM_MEMORY},
				      {CHUNK2_START, CHUNK2_SIZE, RAM_MEMORY},
			      }
};
static struct addrmap *xen_memmap_p = &map4;

/*
 * Initial thread structure.
 *
 * We need to make sure that this is 8192-byte aligned due to the
 * way process stacks are handled. This is done by making sure
 * the linker maps this in the .text segment right after head.S,
 * and making head.S ensure the proper alignment.
 *
 * The things we do for performance..
 */
union thread_union init_thread_union
	__attribute__((__section__(".data.init_task"), __aligned__(THREAD_SIZE))) =
		{ INIT_THREAD_INFO(init_task) };

unsigned long cpu_frequency;


/* opt_dom0_vcpus_pin: If true, dom0 VCPUs are pinned. */
unsigned int opt_dom0_hvm = 1;
boolean_param("dom0_hvm", opt_dom0_hvm);


#define XENHEAP_SIZE (16UL << 20)

static paddr_t xenheap_start;
paddr_t xenheap_end;

static void __init boot_cpu_init(void)
{
	/*
	 * it is assumed that Xen is always booted on cpu0
	 * the logic will have to be changed in multiple
	 * places such as plat_smp_setup() if the assumption
	 * does not hold
	 */
	cpu_set(0, phys_cpu_present_map);
	cpu_set(0, cpu_online_map);

	write_c0_context(processor_id() << (CP0_CONTEXT_PTEBASE_SHIFT + LONGLOG));
}

static int __init parse_xen_memory(const char *p)
{
	unsigned long start = 0, size = 0;
	static int default_map = 1;

	if (default_map) {
		xen_memmap_p->nr_map = 0;
		default_map = 0;
	}

	size = parse_size_and_unit(p, &p);
	if (*p == '@')
		start = parse_size_and_unit(p + 1, &p);
	
	add_addrmap_entry(xen_memmap_p, start, size, RAM_MEMORY);

	return 0;
}
custom_param("xenmem", parse_xen_memory);

#ifdef CONFIG_XEN_UBOOT
uint64_t dom0_loadaddr = 0x08400000;
#else
uint64_t dom0_loadaddr = 0x40400000;
#endif

static int __init set_dom0_loadaddr(const char *p)
{
	dom0_loadaddr = simple_strtoul(p, NULL, 0);

	return 0;
}
custom_param("dom0_loadaddr", set_dom0_loadaddr);

extern unsigned long dom0_size;

static int __init set_dom0_size(const char *p)
{
	dom0_size = simple_strtoul(p, NULL, 0);

	return 0;
}
custom_param("dom0_size", set_dom0_size);

unsigned long dom0_cpumask = 0x1;

static int __init set_dom0_cpumask(const char *p)
{
	dom0_cpumask = simple_strtoul(p, NULL, 0);

	return 0;
}
custom_param("dom0_cpumask", set_dom0_cpumask);

static unsigned int ncores = 1;
static int __init set_ncores(const char *p)
{
	ncores = simple_strtoul(p, NULL, 0);

	return 0;
}
custom_param("ncores", set_ncores);

int dom1_cpu = -1;
integer_param("dom1_cpu", dom1_cpu);

/*
 * We need to store the untouched command line 
 * for future reference. We also need to store 
 * the touched command line since the parameter
 * parsing is performed in place, and we should 
 * allow a component to store reference of 
 * name/value for future reference.
 */

char boot_command_line[COMMAND_LINE_SIZE];
char saved_command_line[COMMAND_LINE_SIZE];
char dom0_command_line[COMMAND_LINE_SIZE];

static char null[4] = { 0 };

static void __init early_cmdline_parse(char **cmdline_p)
{
	char *guest_cmd;
	static const char * const split = "--";

	/*
	 * The format of the command line is: 
	 * "... Xen arguments ..." "[-- dom0 arguments]", i.e., the string
	 * after the splitter "--", if present are meant for dom0
	 */

	if (*cmdline_p == NULL) {
		*cmdline_p = &null[0];
		saved_command_line[0] = '\0';
		dom0_command_line[0] = '\0';
		return;
	}

	guest_cmd = strstr(*cmdline_p, split);

	/* If no spliter, whole line is for guest */
	if (guest_cmd == NULL) {
		guest_cmd = *cmdline_p;
		*cmdline_p = &null[0];
	} else {
		*guest_cmd = '\0';      /* Split boot parameters for xen and guest */
		guest_cmd += strlen(split);
		while (*guest_cmd == ' ') 
			guest_cmd++;
	}

	strlcpy(saved_command_line, *cmdline_p, COMMAND_LINE_SIZE);

	/*
	 * copy the string after the splitter into dom0_command_line
	 */
	strlcpy(dom0_command_line, guest_cmd, COMMAND_LINE_SIZE);
	return;
}

static void __init platform_setup(void)
{
	/*
	 * Place-holder for initializing other platform functionality.
	 * Eg., function pointers such as _machine_restart(), _machine_halt()
	 * and other as the need arises
	 */
}

extern unsigned long max_page;

static void __init bootmem_init(void)
{
	int i;

	for (i = 0; i < xen_memmap_p->nr_map; i++) {
		unsigned long start, end;

		if (xen_memmap_p->map[i].type != RAM_MEMORY)
			continue;

		start = PFN_UP(xen_memmap_p->map[i].addr);
		end = PFN_DOWN(xen_memmap_p->map[i].addr + xen_memmap_p->map[i].size);
		if (end > max_page)
			max_page = end;
#if 0
		if (start < min_low_pfn)
			min_low_pfn = start;
		if (end <= reserved_end)
			continue;
		if (start >= mapstart)
			continue;
		mapstart = max(reserved_end, start);
#endif
	}
	printk("max_page = 0x%lx\n", max_page);
}

static void __init do_initcalls(void)
{
	initcall_t *call;
	for (call = &__initcall_start; call < &__initcall_end; call++)
		(*call)();
}

void __init paging_init(void)
{
	/*
	 * In Linux, zones are configured in this function.
	 * If xen domain allocator is not zone-aware, what
	 * should we do here ? 
	 */
}

static void __init arch_mem_init(char **cmdline_p)
{
	/* 
	 * call board setup routine 
	 */
	platform_setup();

	bootmem_init();
#if 0
	sparse_init();
	paging_init();
#endif
}

static void setup_arch(char **cmdline_p)
{
	cpu_probe();
	/*
	 * the command line arguments passed during bootup
	 * are parsed in cmdline_parse(), while the following
	 * function seperates the arguments for dom0 and
	 * Xen based on the marker "--"
	 */
	early_cmdline_parse(cmdline_p);

	/*
	 * the following call is used to parse arguments to Xen,
	 * the equivalent of Linux's  parse_early_param(). Arguments
	 * to the controlling domain is stored in dom0_command_line[]
	 * Arguments to Xen are string duped into saved_command_line[]
	 */
	cmdline_parse(*cmdline_p);
	
	arch_mem_init(cmdline_p);
}

struct addrmap reserved_memory;

static void reserve_memory(void)
{
	extern char _text[];

	add_addrmap_entry(&reserved_memory, __pa(&_text), xenheap_end - __pa(&_text), RAM_MEMORY);

	printk("Reserved Regions\n");
	printk("---------- -----\n");
	print_addrmap(&reserved_memory, dram_descriptors, sizeof(dram_descriptors));
}

static void init_bootmem_allocator(void)
{
	int i = 0, j = 0;
	__u64 start = 0, end = 0, rsvd_start = 0, rsvd_end = 0;

	for (i = 0; i < xen_memmap_p->nr_map; i++) {

		if (xen_memmap_p->map[i].type != RAM_MEMORY)
			continue;

		start = xen_memmap_p->map[i].addr;
		end = xen_memmap_p->map[i].addr + xen_memmap_p->map[i].size;
				
		for (j = 0; j < reserved_memory.nr_map; j++) {
			rsvd_start = reserved_memory.map[j].addr;
			rsvd_end = rsvd_start + reserved_memory.map[j].size;
			
			if (rsvd_start >= start && rsvd_start < end) {
				if (rsvd_start == start) { /* Continuous exclude */
					start = rsvd_end;
					continue;
				}
				init_boot_pages(start, rsvd_start);
				start = rsvd_end;
			}
			else if ((rsvd_start < start) && (rsvd_end > start))
				start = rsvd_end;
		}

		/* handle the last chunk of memory */
		if (start < end)
			init_boot_pages(start, end);
	}
	
	return;
}

extern char __init_begin[], __init_end[];

static void init_done(void)
{
	memset(__init_begin, 0, __init_end - __init_begin);
	flush_icache_range((unsigned long)__init_begin, (unsigned long)__init_end);
	init_xenheap_pages(__pa(__init_begin), __pa(__init_end));

	startup_cpu_idle_loop();
}

int console_setup(void)
{
	struct ns16550_defaults ns16550 = {
		.data_bits = 8,
		.parity    = 'n',
		.stop_bits = 1
	};

	ns16550.io_base = MMIO_BASEADDR + PHOENIX_IO_UART_0_OFFSET;
	ns16550.irq     = PIC_UART_0_IRQ;
	ns16550.clock_hz = 133000000;
	ns16550.baud = 115200;
	ns16550_init(0, &ns16550);

	console_init_preirq();

	console_force_unlock();

	return 0;
}

extern char reset_entry[], reset_entry_end[];

extern void boot_siblings(void);

/* Configure LSU */
static inline void config_lsu(void)
{
        uint32_t tmp0, tmp1, tmp2;
        __asm__ __volatile__ (
                ".set push\n"
                ".set noreorder\n"
                "li      %0, "STR(LSU_DEFEATURE)"\n"
                "mfcr    %1, %0\n"
                "lui     %2, 0x4080\n"
                "or      %1, %1, %2\n"
                "li      %2, ~0xe\n"
                "and     %1, %1, %2\n"
                "mtcr    %1, %0\n"
		"li      %0, " STR (SCHED_DEFEATURE) "\n"
                "lui     %1, 0x0100\n"
                "mtcr    %1, %0\n"
                ".set pop\n"
                : "=r" (tmp0), "=r" (tmp1), "=r" (tmp2)
        );
}

static void wakeup_secondary(void)
{
	register_t *sys_mmio;
	uint32_t core, value;
	int cpu;
	int i;

	memcpy((void *)NMI_BASE, (void *)&reset_entry, reset_entry_end - reset_entry);

	flush_icache_range((unsigned long)NMI_BASE, (unsigned long)NMI_BASE + reset_entry_end - reset_entry);

	/*
	 * Clear the cpu sibling maps.
	 */
	for (cpu = 0; cpu < NR_CPUS; cpu++)
		cpus_clear(cpu_sibling_map[cpu]);
	cpu_set(0, cpu_sibling_map[0]);

	/* Configure LSU on Core0. */
	config_lsu();

	boot_siblings();

	/* Now boot other cores	 */
	printk("ncores = %d\n", ncores);

	sys_mmio = mmio_address(SYS);
	for (i = 1; i < ncores; i++) {

		core = (0x1 << i);

		/* Enable CPU clock */
		value = read_register(sys_mmio,SYS_COREDFSDISCTRL) & ~core ;
		write_register(sys_mmio, SYS_COREDFSDISCTRL, value);

		/* Remove CPU Reset */
		value = read_register(sys_mmio, SYS_CPURST) & ~core ;
		write_register(sys_mmio, SYS_CPURST, value);

		/* Poll for CPU to mark itself coherent */
		do {
			value = read_register(sys_mmio, SYS_CPUNONCOHERENTMODE) & core;
		} while (value);

		udelay(50);
	}
}

/* ------------------
 * EXTERNAL FUNCTIONS
 * ------------------
 */

struct bootinfo bootinfo;

void (*smp_wakeup)(void *, void *, __u32);

static void build_dom0(void);
static void create_idle_domain(void);

asmlinkage void __start_xen(void)
{
	char *command_line = &boot_command_line[0];

	get_current()->arch.thread_info = &init_thread_info;

	boot_cpu_init();

	console_setup();

	read_bootinfo(&bootinfo);

	setup_arch(&command_line);

	softirq_init();

	/*
	 * init_boot_allocator() returns the physical address 
	 * after the end of the bitmap (1 bit per page frame)
	 * used for book-keeping by the boot memory allocator
	 * The bitmap starts at _end[]
	 */
	xenheap_start = init_boot_allocator(__pa(&_end));

	xenheap_end = xenheap_start + XENHEAP_SIZE; 

	printk("xenheap_start = 0x%lx, xenheap_end = 0x%lx\n", xenheap_start, xenheap_end);

	/*
	 * reserved memory such as Xen's text + data has to
	 * be excluded explicitly. init_bootmem_allocator()
	 * then skips the reserved regions
	 *
	 * TODO: Does the bitmap earmarked in the call to
	 *     init_boot_allocator() get excluded automatically ?
	 *     Or do need to reserve that region too ?
	 */

	reserve_memory();

	/*
	 * remove xen memory region from the board's dram
	 * configuration/memory (passed to Linux). Swapping
	 * is done as the pruned memory map is present in structures
	 * pointed to by the scratch memory map pointer
	 */
	subtract_map(drammap_p, xen_memmap_p, scratchmap_p);
	swap(drammap_p, scratchmap_p);

	printk("Board Dram Map\n");
	printk("---------- \n");
	print_addrmap(drammap_p, dram_descriptors, sizeof(dram_descriptors));

	/*
	 * delete reserved regions from xen's memory.
	 */
	subtract_map(xen_memmap_p, &reserved_memory, scratchmap_p);
	swap(xen_memmap_p, scratchmap_p);

	printk("Xen Memory Map\n");
	printk("---------- \n");
	print_addrmap(xen_memmap_p, dram_descriptors, sizeof(dram_descriptors));

	init_bootmem_allocator();

	init_frametable();

	mm_init();

	/*
	 * TODO: We should also initiazlize Xen heap allocator
	 *       in lowmem area to enable Xen start malloc()ing
	 *       memory for its tables and other data. In x86_64
	 *       Xen heap is integrated into domain heap. Does
	 *       end_boot_allocator() initializes/creates the
	 *       domain allocator ?
	 */
	init_xenheap_pages(xenheap_start, xenheap_end);

	/*
	 * At this point, bootmem allocator is dispensable. The
	 * available memory is reclaimed and allocated to domain
	 * heap allocator. Does domain heap allocator understand
	 * DMA, NORMAL and HIGHMEM zones ? i.e., is domain heap
	 * allocator a zone-allocator ?
	 *	 
	 * Since dom0 kernel will be packaged in Xen's image
	 * we might have to reclaim linux's memory. Since
	 * our kernel image is only ~8MB, we can gloss over
	 * garbage collection for now
	 */
	end_boot_allocator();

	/*
	 * TODO (for virtualization): M2P table needs to be allocated. 
	 *     M2P table captures the mapping between machine frames 
	 *     and the physical frames assigned to domains. This is 
	 *     done in paging_init() in IA-64
	 */

    /* 
	 * Caution: Do not change the order of the following calls 
	 */
	scheduler_init();

	rcu_init();

	/*
	 * PIC needs to be initialized once by the primary cpu and
	 * hence called in the bootup sequence of the primary cpu
	 * as opposed to in the initialization of every cpu
	 */
	pic_init();

	trap_init();

	/* Timer interrupt handling in Xen are broken on silicon! */
	//init_xen_time();

	init_IRQ();

	local_irq_enable();

	wakeup_secondary();

	mmu_init();

	tlb_init();

	cpu_probe();

	udelay(100);

	smp_init();

	create_idle_domain();

	boot_secondary();

	timer_init();

	do_initcalls();

	build_dom0();

	xsm_init(NULL, NULL, 0);

	init_done();

	cpu_halt();
}

static void create_idle_domain(void)
{
	struct domain *idle_domain;

	idle_domain = domain_create(IDLE_DOMAIN_ID, 0, 0);
	if (idle_domain == NULL) {
		printk("idle domain == NULL\n");
		BUG();
	}

	idle_vcpu[0] = get_current();	
	if (alloc_vcpu(idle_domain, 0, 0) == NULL) {
		printk("alloc_vcpu() failed\n");
		BUG();
	}
}

static void build_dom0(void)
{
	int dom0_flags;
	int cpu, vcpu;
	struct vcpu *v;

	dom0_flags = opt_dom0_hvm ? DOMCRF_hvm : 0;

	/* Create domain 0. */
	dom0 = domain_create(0, dom0_flags, DOM0_SSIDREF);
	if (dom0 == NULL || (alloc_vcpu(dom0, 0, 0) == NULL)) {
		panic("Cannot allocate dom0 vcpu 0\n");
		cpu_halt();
	}

	dom0->is_privileged = 1;
	dom0->target = NULL;

	/*
	 * We're going to setup domain0 using the module(s) that we stashed safely
	 * above our heap. The second module, if present, is an initrd ramdisk.
	 *
	 * FIXME: Populate the following four fields from command line arguments
	 *        The arguments should be physical addresses (as opposed to
	 *        virtual addresses). \Partly done
	 *
	 * TODO:  How should Xen handle dom0's ramdisk ?
	 */
	printk("Dom0 Load Address = 0x%lx\n", dom0_loadaddr);

	dom0_memory_start = (unsigned long) __va(dom0_loadaddr);
	dom0_memory_size = -1;		
	dom0_initrd_start = -1;
	dom0_initrd_size = -1;

	/*
	 * Other architecture seem to stash domain0 in the Xen executable
	 * itself through some "modules" mechanims. We currently load it
	 * into physical memory and tell Xen the physical address. Think
	 * about alternatives as we might have to store more than one
	 * module
	 */

	if (construct_dom0(dom0, dom0_memory_start, dom0_memory_size,
					   dom0_initrd_start,dom0_initrd_size, NULL) != 0) {
		printk("Could not set up DOM0 guest OS\n");
		cpu_halt();
	}
	
	for (cpu = ffs(dom0_cpumask), vcpu = 0; cpu < NR_CPUS; ++cpu) {
		if (dom0_cpumask & (1 << cpu)) {
			++vcpu;
			if ((v = alloc_vcpu(dom0, vcpu, cpu)) == NULL) {
				panic("Cannot allocate dom0 vcpu 0\n");
				cpu_halt();
			}
			/*
			 * We do not clear pause_flags of non-zero vcpus
			 * The primary cpu after initialization of required
			 * data triggers the execution of secondary cpus
			 * through a Xen hypercall. The Xen hypercall 
			 * clears the pause_flags of vcpus at which point
			 * they get scheduled onto their respective cpus
			 *
			 * clear_bit(_VPF_down, &v->pause_flags);
			 */
		}
	}

	domain_unpause(dom0);
}

void build_dom1(void)
{
	int dom1_flags;   
	struct domain *dom1;

	dom1_flags = 0;

	dom1 = domain_create(1, dom1_flags, DOMU_SSIDREF);
	if (dom1 == NULL || (alloc_vcpu(dom1, 0, dom1_cpu) == NULL)) {
		panic("Cannot allocate dom1 vcpu 0\n");
		cpu_halt();
	}

	dom1->is_privileged = 1;
	dom1->target = NULL;
        
	if (construct_dom0(dom1, (unsigned long) __va(0x40200000), -1, -1, -1, NULL) != 0) {
		printk("Could not set up DOM1 guest OS\n");
		cpu_halt();
	}

	domain_unpause(dom1);
}

void arch_get_xen_caps(xen_capabilities_info_t *info)
{
	/* Interface name is always xen-3.0-* for Xen-3.x. */
	int major = 3, minor = 0;
	char s[32];

	(*info)[0] = '\0';

	snprintf(s, sizeof(s), "xen-%d.%d-mips ", major, minor);
	safe_strcat(*info, s);	
}

void cpufreq_cmdline_parse(char *str)
{

}
